import json
from dataclasses import asdict, is_dataclass
from datetime import date, datetime


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if is_dataclass(o):
            return asdict(o)
        return super().default(o)


def try_str_to_float(text: str) -> any:
    try:
        return float(text)
    except ValueError:
        return text


def function_exists(cls, name: str) -> bool:
    try:
        return callable(getattr(cls, name))
    except:
        return False



import os
import time
import json
from http import HTTPStatus
from inspect import signature
from dataclasses import dataclass
from typing import List, Tuple, Optional, get_args, get_origin

from ..utils.convert import JSONEncoder
from ..utils.data import import_modules, import_directories
from ..di.data import qualifier_to_data
from ..di.injector import provide
from .data import HTTPRequest, HTTPResponse, http_event_to_request
from .app_controller import AppController


@dataclass()
class _Route:
    func: callable
    request_type: Optional[type]
    dependencies: List[Tuple[type, Optional[str]]]


class AppModule:
    """AWS lambda application main entry point.
    :param controllers: Application controllers.
    :param modules: Modules to force import.
    :param directories: Directories to force import.
    """

    def __init__(self, controllers: List[AppController],
                 modules: Optional[List[str]] = None,
                 directories: Optional[List[str]] = None):

        if modules:
            import_modules(modules)
        elif os.environ.get("IMPORT_MODULES"):
            import_modules(os.environ["IMPORT_MODULES"].split(","))

        root = os.environ["ROOT_DIRECTORY"] if os.environ.get("ROOT_DIRECTORY") else "/var/task/"

        if directories:
            import_directories(root, directories)
        elif os.environ.get("IMPORT_DIRECTORIES"):
            import_directories(root, os.environ["IMPORT_DIRECTORIES"].split(","))

        routes = dict()

        for c in controllers:
            routes.update(c.routes)

        for resource, methods in routes.items():
            for method, handler in methods.items():

                request_type: Optional[type] = None
                dependencies: List[Tuple[type, Optional[str]]] = list()
                data = qualifier_to_data(handler[1]) if handler[1] else dict()

                for key, value in signature(handler[0]).parameters.items():
                    origin = get_origin(value.annotation)
                    cls = origin if origin else value.annotation
                    if cls is HTTPRequest:
                        args = get_args(value.annotation)
                        request_type = args[0] if len(args) > 0 else dict
                        continue
                    dependencies.append((cls, data.get(key)))

                if not request_type:
                    raise Exception("HTTPRequest param is not provided.")
                routes[resource][method] = _Route(handler[0], request_type, dependencies)

        self._routes = routes

    async def serve(self, event: dict, context) -> dict:

        try:
            start = time.time()

            route = self._routes[event["resource"]][event["httpMethod"]]
            if not route:
                raise ValueError(f"unknown route, resource: {event.get('resource')} method: {event.get('httpMethod')}")

            params = [await provide(cls, name) for cls, name in route.dependencies]
            params.append(http_event_to_request(route.request_type, event, context))

            response: HTTPResponse = await route.func(*params)

            status_code = response.status
            body = json.dumps(response.body, cls=JSONEncoder) if response.body else response.body
            headers = response.headers if response.headers else dict()
            headers["X-Process-Time"] = int((time.time() - start) * 1000)

        except Exception as e:
            print(f"error:: {e}")
            status_code = HTTPStatus.BAD_REQUEST
            body = json.dumps({"message": str(e)})
            headers = dict()

        return {
            "statusCode": status_code,
            "body": body,
            "headers": headers
        }



from typing import Optional


class AppController:
    """Module with which you can define endpoints.
    :param base_path: base resource url.
    """

    def __init__(self, base_path: str):
        self.routes = dict()
        self._base_url = base_path

    def route(self, path: str, method: str, qualifier: Optional[str] = None):
        """Function that represents an endpoint.
        :param path: Resource url.
        :param method: HTTP method.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        def _route(func: callable):
            path_ = f"{self._base_url}{path}"
            if path_ not in self.routes:
                self.routes[path_] = dict()
            self.routes[path_][method] = func, qualifier
            return func

        return _route

    def get(self, path: str = "", qualifier: Optional[str] = None):
        """Function that represents an HTTP GET method endpoint.
        :param path: Resource url.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        return self.route(path, "GET", qualifier)

    def post(self, path: str = "", qualifier: Optional[str] = None):
        """Function that represents an HTTP POST method endpoint.
        :param path: Resource url.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        return self.route(path, "POST", qualifier)

    def patch(self, path: str = "", qualifier: Optional[str] = None):
        """Function that represents an HTTP PATCH method endpoint.
        :param path: Resource url.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        return self.route(path, "PATCH", qualifier)

    def delete(self, path: str = "", qualifier: Optional[str] = None):
        """Function that represents an HTTP DELETE method endpoint.
        :param path: Resource url.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        return self.route(path, "DELETE", qualifier)

    def put(self, path: str = "", qualifier: Optional[str] = None):
        """Function that represents an HTTP PUT method endpoint.
        :param path: Resource url.
        :param qualifier: Comma separated string, used by injection package in order to
         specify components by name in case of multiple implementations.
        """

        return self.route(path, "PUT", qualifier)



import glob
import importlib
from os.path import basename, isfile, join

from typing import Optional, List, Dict
from enum import Enum
from dataclasses import dataclass


class OrderType(Enum):
    asc = "ASC"
    desc = "DESC"


@dataclass()
class Order:
    type: OrderType
    alias: str


class ConditionType(Enum):
    any = "ANY"
    range = "RANGE"
    compare = "COMPARE"


@dataclass()
class Condition:
    type: ConditionType
    alias: str
    value: any


@dataclass()
class Paging:
    pageNo: int
    pageSize: int
    totalPages: Optional[int] = None
    totalCount: Optional[int] = None


@dataclass
class Metadata:
    dataElapsed: int


@dataclass()
class Page:
    data: List[Dict[str, any]]
    paging: Paging
    meta: Metadata = None


@dataclass()
class Top:
    data: List[Dict[str, any]]
    topSize: int
    hasMore: bool


@dataclass
class Pageable:
    page_no: int = 0
    page_size: int = 20
    top_size: int = 0


def import_directories(root: str, paths: List[str]):
    for path in paths:
        for module_name in get_module_names(f"{root}{path.replace('.', '/')}"):
            importlib.import_module(f"{path}.{module_name}")


def import_modules(paths: List[str]):
    for path in paths:
        importlib.import_module(path)


def get_module_names(path: str):
    modules = glob.glob(join(path, "*.py"))
    return [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")]



from enum import Enum
from typing import List, Optional
from dataclasses import dataclass

from ..utils.data import Order


class Key:
    def __init__(self, *args):
        self.values = list(args)


class SaveType(Enum):
    insert = "INSERT"
    update = "UPDATE"


class JoinType(Enum):
    inner = ""
    left = "LEFT"
    right = "RIGHT"
    outer = "FULL OUTER"
    left_outer = "LEFT OUTER"
    right_outer = "RIGHT OUTER"
    full_outer = "FULL OUTER"
    cross = "CROSS"
    hash = "HASH"


@dataclass()
class JoinThrough:
    from_column_name: str
    to_column_name: str


@dataclass()
class SimpleColumn:
    name: str
    alias: Optional[str] = None


@dataclass()
class Column:
    name: str
    alias: Optional[str] = None
    updatable: bool = True
    insertable: bool = True


@dataclass()
class StatementField:
    alias: str
    statement: str
    relations_aliases: List[str]


@dataclass()
class Relation:
    table: str
    alias: Optional[str]
    force_join: bool
    columns: List[SimpleColumn]
    join_type: JoinType
    through: JoinThrough


@dataclass()
class Schema:
    table: str
    alias: Optional[str]
    primary_key: List[str]
    order: List[Order]
    statement_fields: List[StatementField]
    columns: List[Column]
    relations: List[Relation]



from typing import List, Optional, TypeVar, Type
from inspect import signature, isfunction, iscoroutinefunction
from .data import Dependency, Scope, qualifier_to_data

T = TypeVar("T")


class _Injector:
    """Module that used in order to resolve dependencies from components or route functions."""

    dependencies: List[Dependency] = list()

    @staticmethod
    def inject(name: Optional[str] = None, scope: Scope = Scope.singleton, qualifier: Optional[str] = None):
        def _inject(component):
            data = qualifier_to_data(qualifier) if qualifier else dict()
            if isfunction(component):
                cls = component.__annotations__.get("return")
                if not cls:
                    raise ValueError(f"injected function '{component.__name__}' must specify return type.")

                params = [(v.annotation, data.get(k)) for k, v in signature(component).parameters.items()]
                _Injector.dependencies.append(Dependency(scope, cls, params, name, factory=component))
            else:
                items_ = dict(filter(lambda i: i[0] != "self", signature(component.__init__).parameters.items()))
                params = [(v.annotation, data.get(k)) for k, v in items_.items()]
                _Injector.dependencies.append(Dependency(scope, component, params, name))
            return component

        return _inject

    @staticmethod
    async def provide(cls: Type[T], name: Optional[str] = None) -> T:
        matched = list(filter(lambda d: (d.cls is cls or issubclass(d.cls, cls)) and
                                        (d.name == name if name else d.name is None), _Injector.dependencies))
        if not matched:
            raise ValueError(f"component name:'{name}', class:'{cls}' not found.")
        if len(matched) > 1:
            raise ValueError(f"multiple matching components have been found for class: '{cls}'")
        if not matched[0].instance:
            nested = list()
            for param in matched[0].params:
                nested.append(await _Injector.provide(param[0], param[1]))  # noqa
            if matched[0].factory:
                instance = await matched[0].factory(*nested) \
                    if iscoroutinefunction(matched[0].factory) else matched[0].factory(*nested)
            else:
                instance = matched[0].cls(*nested)
            if matched[0].scope is Scope.request:
                return instance
            matched[0].instance = instance
        return matched[0].instance

    @staticmethod
    def register_instance(cls: Type[T], instance: T, name: Optional[str] = None):
        _Injector.dependencies.append(Dependency(Scope.singleton, cls, name=name, instance=instance))


def inject(name: Optional[str] = None, scope: Scope = Scope.singleton, qualifier: Optional[str] = None):
    """Inject a class or function.
    :param name: Component name.
    :param scope: Class of the Scope in which to resolve.
    :param qualifier: Comma separated string, used to specify components by name in case of multiple implementations.
    """

    return _Injector.inject(name, scope, qualifier)


async def provide(cls: Type[T], name: Optional[str] = None) -> T:
    """Get an instance of the given type.
    :param cls: Interface whose implementation we want.
    :param name: Name of a specific component that implements the interface.
    :returns: An implementation of interface.
    """

    return await _Injector.provide(cls, name)


def register_instance(cls: Type[T], instance: T, name: Optional[str] = None):
    """Create manually and register an instance of a specific type.
    :param cls: Component type.
    :param instance: Component instance.
    :param name: Component name.
    """

    return _Injector.register_instance(cls, instance, name)



from enum import Enum
from typing import List, Dict, Tuple, Optional, Generic, TypeVar, Type
from dataclasses import dataclass


class Scope(Enum):
    request = "REQUEST"
    singleton = "SINGLETON"


T = TypeVar("T")


@dataclass()
class Dependency(Generic[T]):
    scope: Scope
    cls: Type[T]
    params: List[Tuple[type, Optional[str]]] = None
    name: Optional[str] = None
    instance: Optional[T] = None
    factory: Optional[callable] = None


def qualifier_to_data(qualifier: str) -> Dict[str, str]:
    d = dict()
    for s in qualifier.split(","):
        q = s.split(":")
        d[q[0]] = q[1]
    return d



import json
from dataclasses import dataclass
from typing import Optional, Dict, List, Generic, TypeVar, Type
from http import HTTPStatus

from ..utils.convert import try_str_to_float
from ..utils.data import Order, OrderType, Condition, ConditionType, Pageable

reserved_keys = ["fields", "pageNo", "pageSize", "topSize", "order"]

T = TypeVar("T")


@dataclass()
class HTTPResponse:
    status: HTTPStatus
    body: any = None
    headers: Dict[str, any] = None


@dataclass()
class QueryParameters:
    fields: List[str] = None
    conditions: List[Condition] = None
    page: Pageable = Pageable()
    order: Optional[List[Order]] = None


@dataclass()
class HTTPRequest(Generic[T]):
    query_parameters: QueryParameters = QueryParameters()
    body: Optional[T] = None
    headers: Dict[str, any] = None
    path_parameters: Dict[str, any] = None


def http_event_to_request(cls: Type[T], event: Dict[str, any], context) -> HTTPRequest[T]:
    data = json.loads(event["body"]) if event.get("body") else None

    request = HTTPRequest(QueryParameters(),
                          cls(**data) if cls is not dict else data,
                          event.get("headers"),
                          event.get("pathParameters"))

    params: Optional[Dict[str, any]] = event.get("queryStringParameters")

    if params:
        fields = params["fields"].replace(" ", "").split(",") if params.get("fields") else list()
        pageable = Pageable(page_param(params, alias="pageNo", default=0),
                            page_param(params, alias="pageSize", default=20, max_=50),
                            page_param(params, alias="topSize", default=0, max_=1000))

        conditions = list()
        for k, v in params.items():
            key = k.strip()
            if key in reserved_keys:
                continue
            else:
                value = v.strip()
                value_ = value.lower()
                if value_.startswith("range(") and value_.endswith(")"):
                    condition_type = ConditionType.range
                    data = condition_param("range", value)
                    if not len(data) >= 2:
                        continue
                elif value_.startswith("any(") and value_.endswith(")"):
                    condition_type = ConditionType.any
                    data = condition_param("any", value)
                else:
                    condition_type = ConditionType.compare
                    data = try_str_to_float(value)
                conditions.append(Condition(condition_type, key, data))

        order = list()
        for order_ in (params["order"].split(",") if params.get("order") else list()):
            o = order_.strip().split(" ")
            if len(o) < 1:
                continue
            order_type = OrderType.asc
            if len(o) == 2:
                order_type = next((e for e in OrderType if e.value == o[1].upper()), None)
            if not order_type:
                continue
            order.append(Order(order_type, alias=o[0]))

        request.query_parameters = QueryParameters(fields, conditions, pageable, order)

    return request


def page_param(params: Dict[str, any], alias: str, default: int, max_: Optional[int] = None) -> int:
    try:
        if not params.get(alias):
            return default
        value_ = int(params[alias])
        return value_ if (max_ >= value_ > 0 if max_ else value_ > 0) else default
    except ValueError:
        return default


def condition_param(type_, p) -> list:
    return [try_str_to_float(i) for i in p[len(type_) + 1:len(p) - 1].replace(" ", "").split(",")]
